@@ -1,7 +1,7 @@
-/* @(#)buffer.c	1.28 97/12/06 Copyright 1985, 1995 J. Schilling */
+/* @(#)buffer.c	1.46 01/04/07 Copyright 1985, 1995 J. Schilling */
 #ifndef lint
 static	char sccsid[] =
-	"@(#)buffer.c	1.28 97/12/06 Copyright 1985, 1995 J. Schilling";
+	"@(#)buffer.c	1.46 01/04/07 Copyright 1985, 1995 J. Schilling";
 #endif
 /*
  *	Buffer handling routines
@@ -26,12 +26,10 @@ static	char sccsid[] =
 
 #include <mconfig.h>
 #include <stdio.h>
-#ifdef	HAVE_STDARG_H
-#include <stdarg.h>
-#else
-#include <varargs.h>
-#endif
-#include <timedefs.h>
+#include <fctldefs.h>
+#include <sys/types.h>
+#include <sys/ioctl.h>
+#include <vadefs.h>
 #include "star.h"
 #include <errno.h>
 #include <standard.h>
@@ -40,19 +38,27 @@ static	char sccsid[] =
 #include <unixstd.h>
 #include <strdefs.h>
 #include <waitdefs.h>
+#include <schily.h>
+#ifdef	HAVE_SYS_MTIO_H
+#include <sys/mtio.h>
+#else
+#include "mtio.h"
+#endif
 #include "starsubs.h"
+#include "remote.h"
 
-int	bigcnt	= 0;
+long	bigcnt	= 0;
 int	bigsize	= 0;		/* Tape block size */
 int	bufsize	= 0;		/* Available buffer size */
 char	*bigbuf	= NULL;
 char	*bigptr	= NULL;
+char	*eofptr	= NULL;
 
 m_stats	bstat;
 m_stats	*stats	= &bstat;
 int	pid;
 
-#ifdef	BSD4_2
+#ifdef	timerclear
 LOCAL	struct	timeval	starttime;
 LOCAL	struct	timeval	stoptime;
 #endif
@@ -60,7 +66,7 @@ LOCAL	struct	timeval	stoptime;
 LOCAL	BOOL	isremote = FALSE;
 LOCAL	int	remfd	= -1;
 LOCAL	char	*remfn;
-LOCAL	char	host[64];
+LOCAL	char	host[128];
 
 extern	FILE	*tarf;
 extern	FILE	*tty;
@@ -73,11 +79,15 @@ extern	BOOL	showtime;
 extern	BOOL	no_stats;
 extern	BOOL	do_fifostats;
 extern	BOOL	cflag;
+extern	BOOL	uflag;
+extern	BOOL	rflag;
 extern	BOOL	zflag;
+extern	BOOL	bzflag;
 extern	BOOL	multblk;
 extern	BOOL	partial;
 extern	BOOL	wready;
 extern	BOOL	nullout;
+extern	BOOL	nowarn;
 
 extern	int	intr;
 
@@ -87,6 +97,7 @@ EXPORT	void	closetape	__PR((void));
 EXPORT	void	changetape	__PR((void));
 EXPORT	void	nexttape	__PR((void));
 EXPORT	void	initbuf		__PR((int nblocks));
+EXPORT	void	markeof		__PR((void));
 EXPORT	void	syncbuf		__PR((void));
 EXPORT	int	readblock	__PR((char* buf));
 LOCAL	int	readtblock	__PR((char* buf, int amount));
@@ -107,11 +118,16 @@ EXPORT	void	buf_wake	__PR((int amount));
 EXPORT	int	buf_rwait	__PR((int amount));
 EXPORT	void	buf_rwake	__PR((int amount));
 EXPORT	void	buf_resume	__PR((void));
+EXPORT	void	backtape	__PR((void));
+EXPORT	int	mtioctl		__PR((int cmd, int count));
+EXPORT	long	mtseek		__PR((long offset, int whence));
 EXPORT	int	tblocks		__PR((void));
 EXPORT	void	prstats		__PR((void));
+EXPORT	BOOL	checkerrs	__PR((void));
 EXPORT	void	exprstats	__PR((int ret));
-EXPORT	void	excomerrno	__PR((int err, char* fmt, ...));
-EXPORT	void	excomerr	__PR((char* fmt, ...));
+EXPORT	void	excomerrno	__PR((int err, char* fmt, ...)) __printflike__(2, 3);
+EXPORT	void	excomerr	__PR((char* fmt, ...)) __printflike__(1, 2);
+EXPORT	void	die		__PR((int err));
 LOCAL	void	compressopen	__PR((void));
 
 EXPORT BOOL
@@ -119,22 +135,29 @@ openremote()
 {
 	register char *hp;
 	register char *fp;
+	register int  i;
 
-	if (!nullout && strchr(tarfile, ':')) {
+	if ((!nullout || (uflag || rflag)) && strchr(tarfile, ':')) {
 
+#ifdef	USE_REMOTE
 		isremote = TRUE;
 		remfn = strchr(tarfile, ':');
-		for (fp = tarfile, hp = host; fp < remfn;)
+		for (fp = tarfile, hp = host, i = 1;
+				fp < remfn && i < sizeof(host); i++) {
 			*hp++ = *fp++;
+		}
 		*hp = '\0';
 		remfn++;
 		if (debug)
-			errmsgno(BAD, "Remote: %s Host: %s file: %s\n",
+			errmsgno(EX_BAD, "Remote: %s Host: %s file: %s\n",
 							tarfile, host, remfn);
 
 		if ((remfd = rmtgetconn(host, bigsize)) < 0)
-			comerrno(BAD, "Cannot get connection to '%s'.\n",
+			comerrno(EX_BAD, "Cannot get connection to '%s'.\n",
 				/* errno not valid !! */		host);
+#else
+		comerrno(EX_BAD, "Remote tape support not present.\n");
+#endif
 	}
 	return (isremote);
 }
@@ -144,7 +167,7 @@ opentape()
 {
 	int	n = 0;
 
-	if (nullout) {
+	if (nullout && !(uflag || rflag)) {
 		tarfile = "null";
 		tarf = (FILE *)NULL;
 	} else if (streql(tarfile, "-")) {
@@ -157,28 +180,65 @@ opentape()
 			multblk = TRUE;
 		}
 		setbuf(tarf, (char *)NULL);
+#if	defined(__CYGWIN32__)
+		setmode(fileno(tarf), O_BINARY);
+#endif
 	} else if (isremote) {
-		while (rmtopen(remfd, remfn, cflag ? 2:0) < 0) {
+#ifdef	USE_REMOTE
+		/*
+		 * isremote will always be FALSE if USE_REMOTE is not defined.
+		 * NOTE any open flag bejond O_RDWR is not portable across
+		 * different platforms. The remote tape library will check
+		 * whether the current /etc/rmt server supports symbolic
+		 * open flags. If there is no symbolic support in the
+		 * remote server, our rmt client code will mask off all
+		 * non portable bits.
+		 */
+		while (rmtopen(remfd, remfn, (cflag ? O_RDWR|O_CREAT:O_RDONLY)|O_BINARY) < 0) {
 			if (!wready || n++ > 6 || geterrno() != EIO)
 				comerr("Cannot open remote '%s'.\n", tarfile);
 			else
 				sleep(10);
 		}
-		
-	} else while ((tarf = fileopen(tarfile, cflag?"rwcu":"ru")) ==
+#endif	
+	} else {
+		FINFO	finfo;
+	extern	BOOL	follow;
+	extern	BOOL	fcompat;
+
+		if (fcompat && cflag) {
+			/*
+			 * The old syntax has a high risk of corrupting
+			 * files if the user disorders the args.
+			 * For this reason, we do not allow to overwrite
+			 * a plain file in compat mode.
+			 * XXX What if we implement 'r' & 'u' ???
+			 */
+			follow++;
+			n = getinfo(tarfile, &finfo);
+			follow--;
+			if (n >= 0 && is_file(&finfo) && finfo.f_size > 0L) {
+				comerrno(EX_BAD,
+				"Will not overwrite non empty plain files in compat mode.\n");
+			}
+		}
+
+		n = 0;
+		while ((tarf = fileopen(tarfile, cflag?"rwcub":"rub")) ==
 								(FILE *)NULL) {
-		if (!wready || n++ > 6 || geterrno() != EIO)
-			comerr("Cannot open '%s'.\n", tarfile);
-		else
-			sleep(10);
+			if (!wready || n++ > 6 || geterrno() != EIO)
+				comerr("Cannot open '%s'.\n", tarfile);
+			else
+				sleep(10);
+		}
 	}
-	if (!isremote && !nullout) {
+	if (!isremote && (!nullout || (uflag || rflag))) {
 		file_raise(tarf, FALSE);
 		checkarch(tarf);
 	}
 	vpr = tarf == stdout ? stderr : stdout;
 
-	if (zflag) {
+	if (zflag || bzflag) {
 		extern	long	tape_dev;
 		extern	long	tape_ino;
 
@@ -190,7 +250,7 @@ opentape()
 			comerrno(EX_BAD, "Can only compress files.\n");
 	}
 
-#ifdef	BSD4_2
+#ifdef	timerclear
 	if (showtime && gettimeofday(&starttime, (struct timezone *)0) < 0)
 		comerr("Cannot get starttime\n");
 #endif
@@ -200,8 +260,13 @@ EXPORT void
 closetape()
 {
 	if (isremote) {
+#ifdef	USE_REMOTE
+		/*
+		 * isremote will always be FALSE if USE_REMOTE is not defined.
+		 */
 		if (rmtclose(remfd) < 0)
 			errmsg("Remote close failed.\n");
+#endif
 	} else {
 		if (tarf)
 			fclose(tarf);
@@ -219,7 +284,7 @@ changetape()
 	stats->blocks = 0L;
 	stats->parts = 0L;
 	closetape();
-	errmsgno(BAD, "Mount volume #%d and hit <RETURN>", ++stats->volno);
+	errmsgno(EX_BAD, "Mount volume #%d and hit <RETURN>", ++stats->volno);
 	fgetline(tty, ans, sizeof(ans));
 	if (feof(tty))
 		exit(1);
@@ -261,11 +326,38 @@ initbuf(nblocks)
 	stats->swapflg = -1;
 }
 
+EXPORT void
+markeof()
+{
+#ifdef	FIFO
+	if (use_fifo) {
+		/*
+		 * Remember current FIFO status.
+		 */
+	}
+#endif
+	eofptr = bigptr - TBLOCK;
+
+	if (debug) {
+		error("Blocks: %d\n", tblocks());
+		error("bigptr - bigbuff: %d %p %p %p lastsize: %ld\n",
+			bigptr - bigbuf, bigbuf, bigptr, eofptr, stats->lastsize);
+	}
+}
+
 EXPORT void
 syncbuf()
 {
-	bigptr = bigbuf;
-	fillbytes(bigbuf, bigsize, '\0');
+#ifdef	FIFO
+	if (use_fifo) {
+		/*
+		 * Switch FIFO direction.
+		 */
+		excomerr("Cannot update tape with FIFO.\n");
+	}
+#endif
+	bigptr = eofptr;
+	bigcnt = eofptr - bigbuf;
 }
 
 EXPORT int
@@ -288,8 +380,13 @@ readtblock(buf, amount)
 
 	stats->reading = TRUE;
 	if (isremote) {
+#ifdef	USE_REMOTE
+		/*
+		 * isremote will always be FALSE if USE_REMOTE is not defined.
+		 */
 		if ((cnt = rmtread(remfd, buf, amount)) < 0)
 			excomerr("Error reading '%s'.\n", tarfile);
+#endif
 	} else {
 		if ((cnt = ffileread(tarf, buf, amount)) < 0)
 			excomerr("Error reading '%s'.\n", tarfile);
@@ -329,21 +426,21 @@ readtape(buf, amount)
 	if (amt == 0)
 		return (amt);
 	if (amt < TBLOCK)
-		excomerrno(BAD, "Error reading '%s' size (%d) too small.\n",
+		excomerrno(EX_BAD, "Error reading '%s' size (%d) too small.\n",
 							tarfile, amt);
 	/*
 	 * First block
 	 */
 	if (stats->swapflg < 0) {
 		if ((amt % TBLOCK) != 0)
-			comerrno(BAD, "Invalid blocksize %d bytes.\n", amt);
+			comerrno(EX_BAD, "Invalid blocksize %d bytes.\n", amt);
 		if (amt < amount) {
 			stats->blocksize = bigsize = amt;
 #ifdef	FIFO
 			if (use_fifo)
 				fifo_ibs_shrink(amt);
 #endif
-			errmsgno(BAD, "Blocksize = %d records.\n",
+			errmsgno(EX_BAD, "Blocksize = %ld records.\n",
 						stats->blocksize/TBLOCK);
 		}
 	}
@@ -354,6 +451,7 @@ readtape(buf, amount)
 		stats->blocks++;
 	else
 		stats->parts += amt;
+	stats->lastsize = amt;
 #ifdef	DEBUG
 	error("readbuf: cnt: %d.\n", amt);
 #endif
@@ -410,18 +508,27 @@ writetape(buf, amount)
 	int	amount;
 {
 	int	cnt;
+	int	err = 0;
 					/* hartes oder weiches EOF ??? */
 					/* d.h. < 0 oder <= 0          */
 	stats->reading = FALSE;
 	if (nullout) {
 		cnt = amount;
+#ifdef	USE_REMOTE
 	} else if (isremote) {
-		if ((cnt = rmtwrite(remfd, buf, amount)) <= 0)
-			excomerr("Error writing '%s'.\n", tarfile);
+		cnt = rmtwrite(remfd, buf, amount);
+#endif
 	} else {
-		if ((cnt = ffilewrite(tarf, buf, amount)) <= 0) /* XXX */
-			excomerr("Error writing '%s'.\n", tarfile);
+		cnt = ffilewrite(tarf, buf, amount);
 	}
+	if (cnt == 0) {
+		err = EFBIG;
+	} else if (cnt < 0) {
+		err = geterrno();
+	}
+	if (cnt <= 0)
+		excomerrno(err, "Error writing '%s'.\n", tarfile);
+
 	if (cnt == stats->blocksize)
 		stats->blocks++;
 	else
@@ -574,6 +681,81 @@ buf_resume()
 #endif
 }
 
+EXPORT void
+backtape()
+{
+	long	ret;
+
+	if (debug) {
+		error("Blocks: %d\n", tblocks());
+		error("filepos: %ld seeking to: %ld bigsize: %d\n",
+		mtseek(0L, SEEK_CUR), mtseek(0L, SEEK_CUR)-stats->lastsize, bigsize);
+	}
+
+	if (mtioctl(MTNOP, 0) >= 0) {
+		if (debug)
+			error("Is a tape: BSR 1...\n");
+		ret = mtioctl(MTBSR, 1);
+	} else {
+		if (debug)
+			error("Is a file: lseek()\n");
+		ret = mtseek(-stats->lastsize, SEEK_CUR);
+	}
+	if (ret < 0)
+		excomerr("Cannot backspace tape.\n");
+
+	if (stats->lastsize == stats->blocksize)
+		stats->blocks--;
+	else
+		stats->parts -= stats->lastsize;
+}
+
+EXPORT int
+mtioctl(cmd, count)
+	int	cmd;
+	int	count;
+{
+	int	ret;
+
+	if (nullout && !(uflag || rflag)) {
+		return (0);
+#ifdef	USE_REMOTE
+	} else if (isremote) {
+		ret = rmtioctl(remfd, cmd, count);
+#endif
+	} else {
+#ifdef	MTIOCTOP
+		struct mtop mtop;
+
+		mtop.mt_op = cmd;
+		mtop.mt_count = count;
+
+		ret = ioctl(fdown(tarf), MTIOCTOP, &mtop);
+#else
+		return (-1);
+#endif
+	}
+	if (ret < 0 && debug)
+		errmsg("Error sending mtioctl(%d, %d) to '%s'.\n", cmd, count, tarfile);
+	return (ret);
+}
+
+EXPORT long
+mtseek(offset, whence)
+	long	offset;
+	int	whence;
+{
+	if (nullout && !(uflag || rflag)) {
+		return (0L);
+#ifdef	USE_REMOTE
+	} else if (isremote) {
+		return (rmtseek(remfd, offset, whence));
+#endif
+	} else {
+		return (fseek(tarf, offset, whence));
+	}
+}
+
 EXPORT int
 tblocks()
 {
@@ -584,7 +766,7 @@ tblocks()
 		fifo_cnt = fifo_amount()/TBLOCK;
 #endif
 	if (debug) {
-		error("blocks: %d blocksize: %d parts: %d bigcnt: %d fifo_cnt: %d\n", 
+		error("blocks: %ld blocksize: %ld parts: %ld bigcnt: %ld fifo_cnt: %ld\n", 
 		stats->blocks, stats->blocksize, stats->parts, bigcnt, fifo_cnt);
 	}
 	if (stats->reading)
@@ -604,7 +786,7 @@ prstats()
 	long	lobytes;
 	char	cbytes[32];
 	int	per;
-#ifdef	BSD4_2
+#ifdef	timerclear
 	int	sec;
 	int	usec;
 	int	tmsec;
@@ -615,7 +797,7 @@ prstats()
 	if (pid == 0)	/* child */
 		return;
 
-#ifdef	BSD4_2
+#ifdef	timerclear
 	if (showtime && gettimeofday(&stoptime, (struct timezone *)0) < 0)
 		comerr("Cannot get stoptime\n");
 #endif
@@ -636,8 +818,8 @@ prstats()
 	else
 		sprintf(cbytes, "%ld", lobytes);
 
-	errmsgno(BAD,
-		"%l blocks + %l bytes (total of %s bytes = %l.%02dk).\n",
+	errmsgno(EX_BAD,
+		"%ld blocks + %ld bytes (total of %s bytes = %ld.%02dk).\n",
 		stats->blocks, stats->parts, cbytes, kbytes, per);
 
 	if (stats->Tblocks + stats->Tparts) {
@@ -654,11 +836,11 @@ prstats()
 		else
 			sprintf(cbytes, "%ld", lobytes);
 
-		errmsgno(BAD,
-		"Total %l blocks + %l bytes (total of %s bytes = %l.%02dk).\n",
+		errmsgno(EX_BAD,
+		"Total %ld blocks + %ld bytes (total of %s bytes = %ld.%02dk).\n",
 		stats->Tblocks, stats->Tparts, cbytes, kbytes, per);
 	}
-#ifdef	BSD4_2
+#ifdef	timerclear
 	if (showtime) {
 		Llong	kbs;
 
@@ -675,12 +857,42 @@ prstats()
 		kbs = (Llong)kbytes*(Llong)1000/tmsec;
 		lobytes = kbs;
 
-		errmsgno(BAD, "Total time %d.%03dsec (%ld kBytes/sec)\n",
+		errmsgno(EX_BAD, "Total time %d.%03dsec (%ld kBytes/sec)\n",
 				sec, usec/1000, lobytes);
 	}
 #endif
 }
 
+EXPORT BOOL
+checkerrs()
+{
+	if (xstats.s_staterrs	||
+	    xstats.s_openerrs	||
+	    xstats.s_rwerrs	||
+	    xstats.s_misslinks	||
+	    xstats.s_toolong	||
+	    xstats.s_toobig	||
+	    xstats.s_isspecial	||
+	    xstats.s_sizeerrs) {
+		if (nowarn || no_stats || (pid == 0)/* child */)
+			return (TRUE);
+
+		errmsgno(EX_BAD, "The following problems occurred during archive processing:\n");
+		errmsgno(EX_BAD, "Cannot: stat %d, open %d, read/write %d. Size changed %d.\n",
+				xstats.s_staterrs,
+				xstats.s_openerrs,
+				xstats.s_rwerrs,
+				xstats.s_sizeerrs);
+		errmsgno(EX_BAD, "Missing links %d, Name too long %d, File too big %d, Not dumped %d.\n",
+				xstats.s_misslinks,
+				xstats.s_toolong,
+				xstats.s_toobig,
+				xstats.s_isspecial);
+		return (TRUE);
+	}
+	return (FALSE);
+}
+
 EXPORT void
 exprstats(ret)
 	int	ret;
@@ -710,6 +922,9 @@ excomerrno(err, fmt, va_alist)
 #endif
 	errmsgno(err, "%r", fmt, args);
 	va_end(args);
+#ifdef	FIFO
+	fifo_exit();
+#endif
 	exprstats(err);
 	/* NOTREACHED */
 }
@@ -735,10 +950,20 @@ excomerr(fmt, va_alist)
 #endif
 	errmsgno(err, "%r", fmt, args);
 	va_end(args);
+#ifdef	FIFO
+	fifo_exit();
+#endif
 	exprstats(err);
 	/* NOTREACHED */
 }
 
+EXPORT void
+die(err)
+	int	err;
+{
+	excomerrno(err, "Cannot recover from error - exiting.\n");
+}
+
 /*
  * Quick hack to implement a -z flag. May be changed soon.
  */
@@ -751,6 +976,10 @@ compressopen()
 {
 	FILE	*pp[2];
 	int	mypid;
+	char	*zip_prog = "gzip";
+        
+	if (bzflag)
+		zip_prog = "bzip2";
 
 	multblk = TRUE;
 
@@ -769,14 +998,21 @@ compressopen()
 		else
 			fclose(pp[0]);
 
+#if	defined(__CYGWIN32__)
+		if (cflag)
+			setmode(fileno(pp[0]), O_BINARY);
+		else
+			setmode(fileno(pp[1]), O_BINARY);
+#endif
+
 		/* We don't want to see errors */
 		null = fileopen("/dev/null", "rw");
 
 		if (cflag)
-			fexecl("gzip", pp[0], tarf, null, "gzip", flg, NULL);
+			fexecl(zip_prog, pp[0], tarf, null, zip_prog, flg, NULL);
 		else
-			fexecl("gzip", tarf, pp[1], null, "gzip", "-d", NULL);
-		errmsg("Compress exec failed\n");
+			fexecl(zip_prog, tarf, pp[1], null, zip_prog, "-d", NULL);
+		errmsg("Compress: exec of '%s' failed\n", zip_prog);
 		_exit(-1);
 	}
 	fclose(tarf);
